﻿using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;

namespace AT
{
	namespace UBF
	{
		namespace Behaviors
		{
			/// <summary>
			/// Turns the turret until directly looking at the enemy. 
			/// If there is not an enemy but there is bullets, then it points the turret at the bullet.
			/// This can be used to evade bullets that aren't exactly in between the two tanks
			/// </summary>
			public class DecentAgent_FindEnemyAndBullets : AT.UBF.Behavior
			{
				/// <summary>
				/// Priority of this behavior when there isn't an enemy in sight
				/// </summary>
				public float priorityWhenNoEnemyInSight = 1.0f;

				/// <summary>
				/// Priority when there is an enemy in sight
				/// </summary>
				public float priorityWhenEnemyInSight = 0.1f;

				/// <summary>
				/// Speed to turn the turret when there isn't an enemy in sight
				/// </summary>
				public float turretTurnSpeedWhenNoEnemyInSight = 1.0f;

				/// <summary>
				/// Determines if this agent is the master or the slave
				/// </summary>
				public bool master = false;

				/// <summary>
				/// Draw the debugging lines to show what is going on inside the behavior
				/// </summary>
				public bool debuggingLines = false;

				/// <summary>
				/// Generate the action associated with the FindEnemyAndBullets behavior
				/// </summary>
				/// <param name="state">State of the world as seen from the tank</param>
				/// <returns>Action that corresponds to this behavior for the given state</returns>
				protected override AT.UBF.Action genActionInternal(AT.UBF.State state)
				{
					AT.UBF.Action action = new AT.UBF.Action();

					List<string> originalMessages;
					var cstate = DecentAgentUtil.GetCombinedStateFromMessages(state, out originalMessages);

					//Master determines who should rotate which direction
					if (master)
					{
						if (cstate.teamates.Count > 0)
						{
							//Sort the teamates to ensure that they get assigned the same turret direction every time so that they don't switch at every timestep somehow
							cstate.teamates.Sort((x, y) => x.color.CompareTo(y.color));
							for (int i = 0; i < cstate.teamates.Count; i++)
							{
								if (i % 2 == 0)
								{
									DecentAgentUtil.SendNonStateMesssage(state, voteSourceID, "Slave: " + this.GetType().Name + ", RotateLeft: " + cstate.teamates[i].color);
								}
								else
								{
									DecentAgentUtil.SendNonStateMesssage(state, voteSourceID, "Slave: " + this.GetType().Name + ", RotateRight: " + cstate.teamates[i].color);
								}
							}
						}
					}
					//If all teamates have been killed, then become master
					else if (cstate.teamates.Count == 0)
					{
						master = true;
					}

					//If there are no visible tanks or bullets, then rotate the turret
					if (cstate.combinedVisibleEnemyTanks.Count < 1 && cstate.combinedVisibleBullets.Count < 1)
					{
						if (originalMessages != null)
						{
							//Determine wether to rotate right or left from messages and master status
							if (getRotateDirection(originalMessages, state.self.getColor()))
							{ action.SetTurnTurretRight(turretTurnSpeedWhenNoEnemyInSight, voteSourceID); }
							else
							{ action.SetTurnTurretRight(-turretTurnSpeedWhenNoEnemyInSight, voteSourceID); }
						}
						else
						{
							action.SetTurnTurretRight(turretTurnSpeedWhenNoEnemyInSight, voteSourceID);
						}
						action.SetPriority(priorityWhenNoEnemyInSight, voteSourceID);
					}
					//If there are any visible enemy tanks, lock on to them
					else if (cstate.combinedVisibleEnemyTanks.Count >= 1)
					{
						var closestTank = DecentAgentUtil.getClosestEnemyTank(state, cstate.combinedVisibleEnemyTanks, false);
						if (debuggingLines)
						{
							Debug.DrawLine(state.self.getPosition(), closestTank.position, Color.red);
						}
						// Pure pursuit (no lag or lead on turret)
						action.SetTurnTurretRight(
							Util.findShortestTurretRotationToTarget(state.self.getPosition(), closestTank.position, state.self.getTurretRotation()),
							voteSourceID);
						action.SetPriority(priorityWhenEnemyInSight, voteSourceID);
						DecentAgentUtil.SendNonStateMesssage(state, voteSourceID, this.GetType().Name + ": " + state.self.getColor().ToString() + " pointing turret at: " + closestTank.color.ToString());
					}
					else
					{
						if (debuggingLines)
						{
							Debug.DrawLine(state.self.getPosition(), 100 * new Vector2((float)Math.Cos(state.self.getTurretRotation().DegreesToRadians()), (float)Math.Sin(state.self.getTurretRotation().DegreesToRadians())), Color.yellow);
						}
						var bullet = DecentAgentUtil.getClosestIncomingBullet(state.self.getPosition(), cstate.combinedVisibleBullets);

						//If all the bullets are outgoing, then continue with the find enemy tank action
						if (bullet != null)
						{
							// Pure pursuit (no lag or lead on turret)
							action.SetTurnTurretRight(
								Util.findShortestTurretRotationToTarget(state.self.getPosition(), bullet.first, state.self.getTurretRotation()),
								voteSourceID);
							action.SetPriority(priorityWhenEnemyInSight, voteSourceID);
						}
						else
						{
							if (originalMessages != null)
							{
								//Get the direction to rotate
								if (getRotateDirection(originalMessages, state.self.getColor()))
								{ action.SetTurnTurretRight(turretTurnSpeedWhenNoEnemyInSight, voteSourceID); }
								else
								{ action.SetTurnTurretRight(-turretTurnSpeedWhenNoEnemyInSight, voteSourceID); }
							}
							else
							{
								action.SetTurnTurretRight(turretTurnSpeedWhenNoEnemyInSight, voteSourceID);
							}
							action.SetPriority(priorityWhenNoEnemyInSight, voteSourceID);
						}
					}

					return action;
				}

				/// <summary>
				/// Gets the direction to rotate. True means to rotate right, and false means to rotate left
				/// </summary>
				/// <param name="messages">List of messages to parse looking for the direction to turn</param>
				/// <param name="color">Color of the current tank</param>
				/// <returns>True if the tank should rotate right, false if left</returns>
				private bool getRotateDirection(List<string> messages, string color)
				{
					//Master always rotates right when no enemies present
					if (master)
					{
						return true;
					}

					//If a slave, then we have to look through all the messages to check if there is one for us
					foreach (var message in messages)
					{
						if (message.Contains(this.GetType().Name))
						{
							if (message.Contains(color))
							{
								if (message.Contains("RotateRight"))
								{
									return true;
								}
								else
								{
									return false;
								}
							}
						}
					}
					return true;
				}

				/** Save arguments to dictionary.
                 */
				public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction)
				{
					dict["PriorityWhenNoEnemyInSight"] = priorityWhenNoEnemyInSight;
					dict["PriorityWhenEnemyInSight"] = priorityWhenEnemyInSight;
					dict["TurretTurnSpeedWhenNoEnemyInSight"] = turretTurnSpeedWhenNoEnemyInSight;
					dict["Master"] = master;
					dict["DebuggingLines"] = debuggingLines;
				}

				/** Loads arguments from a dictionary, verifying as needed.
                 */
				public override void _FromDict(Dictionary<string, object> dict)
				{
					string bName = this.GetType().Name;
					priorityWhenNoEnemyInSight = Util.FloatFromJsonDict("PriorityWhenNoEnemyInSight", priorityWhenNoEnemyInSight, bName, dict);
					priorityWhenEnemyInSight = Util.FloatFromJsonDict("PriorityWhenEnemyInSight", priorityWhenEnemyInSight, bName, dict);
					turretTurnSpeedWhenNoEnemyInSight = Util.FloatFromJsonDict("TurretTurnSpeedWhenNoEnemyInSight", turretTurnSpeedWhenNoEnemyInSight, bName, dict);
					master = DecentAgentUtil.BoolFromJsonDict("Master", master, bName, dict);
					debuggingLines = DecentAgentUtil.BoolFromJsonDict("DebuggingLines", debuggingLines, bName, dict);
				}
			}
		}
	}
}